# Generated by Django 2.2.24 on 2021-07-08 12:40

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0069_update_json_fields'),
    ]

    operations = [
        migrations.RenameField(
            model_name='repository',
            old_name='retained_versions',
            new_name='retain_repo_versions',
        ),
    ]
