# Generated by Django 2.2.5 on 2019-09-13 21:23

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0005_progressreport_code'),
    ]

    operations = [
        migrations.AddField(
            model_name='repository',
            name='plugin_managed',
            field=models.BooleanField(default=False),
        ),
    ]
