# Generated by Django 2.2.14 on 2020-08-25 13:40

import django.contrib.postgres.fields.jsonb
from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0042_rbac_for_tasks'),
    ]

    operations = [
        migrations.AddField(
            model_name='pulpexport',
            name='toc_info',
            field=django.contrib.postgres.fields.jsonb.JSONField(null=True),
        ),
    ]
