# Generated by Django 2.2.13 on 2020-06-30 14:36

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0034_groupprogressreport'),
    ]

    operations = [
        migrations.AddField(
            model_name='content',
            name='upstream_id',
            field=models.UUIDField(null=True),
        ),
    ]
