# Generated by Django 2.2.24 on 2021-06-15 12:11

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0064_repository_user_hidden'),
        ('core', '0064_add_new_style_task_columns'),
    ]

    operations = [
    ]
