# Generated by Django 2.2.7 on 2019-11-25 19:37

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0016_charfield_to_textfield'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='task',
            name='parent',
        ),
    ]
