# Generated by Django 2.2.16 on 2020-10-30 09:31

from django.db import migrations, models
import pulpcore.app.models.upload


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0048_fips_checksums'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='upload',
            name='file',
        ),
        migrations.AddField(
            model_name='uploadchunk',
            name='file',
            field=models.FileField(default=None, max_length=255, upload_to=pulpcore.app.models.upload.UploadChunk.storage_path),
            preserve_default=False,
        ),
    ]
