# Generated by Django 3.2.5 on 2021-08-04 17:49

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0071_filesystemexport_filesystemexporter'),
    ]

    operations = [
        migrations.AddField(
            model_name='filesystemexporter',
            name='method',
            field=models.CharField(choices=[('write', 'Export by writing'), ('hardlink', 'Export by hardlinking'), ('symlink', 'Export by symlinking')], default='write', max_length=128),
        ),
    ]
