# Generated by Django 2.2.16 on 2020-09-25 22:47

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0051_timeoutfields'),
    ]

    operations = [
        migrations.AddField(
            model_name='task',
            name='logging_cid',
            field=models.CharField(db_index=True, default='', max_length=256),
        ),
    ]
