# Generated by Django 3.2.6 on 2021-08-25 20:10

from django.db import migrations, models
import django.db.models.deletion
import pulpcore.app.models.access_policy


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0074_acs'),
    ]

    operations = [
        migrations.CreateModel(
            name='RBACContentGuard',
            fields=[
                ('contentguard_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, related_name='core_rbaccontentguard', serialize=False, to='core.contentguard')),
            ],
            options={
                'permissions': (('download_rbaccontentguard', 'Can Download Content'),),
                'default_related_name': '%(app_label)s_%(model_name)s',
            },
            bases=('core.contentguard', pulpcore.app.models.access_policy.AutoAddObjPermsMixin, pulpcore.app.models.access_policy.AutoDeleteObjPermsMixin),
        ),
    ]
