# Generated by Django 2.2.16 on 2020-09-08 16:32

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0047_improve_orphan_cleanup'),
    ]

    operations = [
        migrations.AlterField(
            model_name='artifact',
            name='md5',
            field=models.CharField(db_index=True, max_length=32, null=True),
        ),
        migrations.AlterField(
            model_name='artifact',
            name='sha1',
            field=models.CharField(db_index=True, max_length=40, null=True),
        ),
        migrations.AlterField(
            model_name='artifact',
            name='sha224',
            field=models.CharField(db_index=True, max_length=56, null=True),
        ),
        migrations.AlterField(
            model_name='artifact',
            name='sha384',
            field=models.CharField(db_index=True, max_length=96, null=True, unique=True),
        ),
        migrations.AlterField(
            model_name='artifact',
            name='sha512',
            field=models.CharField(db_index=True, max_length=128, null=True, unique=True),
        ),
    ]
