# Generated by Django 2.2.15 on 2020-09-09 19:22

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0046_task__resource_job_id'),
    ]

    operations = [
        migrations.AlterField(
            model_name='contentartifact',
            name='artifact',
            field=models.ForeignKey(null=True, on_delete=django.db.models.deletion.PROTECT, related_name='content_memberships', to='core.Artifact'),
        ),
        migrations.AlterField(
            model_name='basedistribution',
            name='pulp_type',
            field=models.TextField(db_index=True, default=None),
        ),
        migrations.AlterField(
            model_name='content',
            name='pulp_type',
            field=models.TextField(db_index=True, default=None),
        ),
        migrations.AlterField(
            model_name='contentguard',
            name='pulp_type',
            field=models.TextField(db_index=True, default=None),
        ),
        migrations.AlterField(
            model_name='exporter',
            name='pulp_type',
            field=models.TextField(db_index=True, default=None),
        ),
        migrations.AlterField(
            model_name='importer',
            name='pulp_type',
            field=models.TextField(db_index=True, default=None),
        ),
        migrations.AlterField(
            model_name='publication',
            name='pulp_type',
            field=models.TextField(db_index=True, default=None),
        ),
        migrations.AlterField(
            model_name='remote',
            name='pulp_type',
            field=models.TextField(db_index=True, default=None),
        ),
        migrations.AlterField(
            model_name='repository',
            name='pulp_type',
            field=models.TextField(db_index=True, default=None),
        ),
    ]
