# Generated by Django 2.2.23 on 2021-06-03 20:05

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0070_rename_retained_versions'),
    ]

    operations = [
        migrations.CreateModel(
            name='FilesystemExport',
            fields=[
                ('export_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='core.Export')),
            ],
            options={
                'abstract': False,
            },
            bases=('core.export',),
        ),
        migrations.CreateModel(
            name='FilesystemExporter',
            fields=[
                ('exporter_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, related_name='core_fs_exporter', serialize=False, to='core.Exporter')),
                ('path', models.TextField()),
            ],
            options={
                'default_related_name': '%(app_label)s_fs_exporter',
            },
            bases=('core.exporter',),
        ),
    ]
