# Generated by Django 2.2.15 on 2020-09-02 16:16

import django.contrib.postgres.fields.jsonb
from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0044_temp_file_artifact_field'),
    ]

    operations = [
        migrations.AlterField(
            model_name='accesspolicy',
            name='permissions_assignment',
            field=django.contrib.postgres.fields.jsonb.JSONField(null=True),
        ),
    ]
