# Generated by Django 2.2.5 on 2019-09-09 20:17

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0004_add_duplicated_reserved_resources'),
    ]

    operations = [
        migrations.AddField(
            model_name='progressreport',
            name='code',
            field=models.CharField(default="no_code", max_length=36),
            preserve_default=False,
        ),
    ]
